import asyncio
import shutil
import math
from pathlib import Path
import concurrent.futures
import multiprocessing as mp
import subprocess
import shlex
from py7zr import unpack_7zarchive
import orjson
import pickle
import numpy as np
import pandas as pd
import requests
import aiofiles
import io
from tqdm.notebook import tqdm
from dl2050utils.core import oget, listify, xre

Path.ls = lambda x: list(x.iterdir())

# ################################################################################
# pickle_save (pickle_dump), pickle_load
# np_save, np_load, np_load_async
# df_save, df_load
# json_load (json_parse), json_dump, read_json
# json_saves, json_loads
# ################################################################################


def pickle_save(p, d):
    """Saves an object to a pickle file.

    Args:
        p (str or Path): The file path where the object will be saved.
        d (any): The object to be saved.

    Returns:
        int: 0 if successful, 1 if an error occurred.
    """
    p = Path(p)
    p = p.with_suffix(".pickle")
    try:
        with open(p, "wb") as f:
            pickle.dump(d, f)
        return 0
    except Exception as exc:
        print(str(exc))
        return 1


# Deprecated
def pickle_dump(p, d):
    return pickle_save(p, d)


def pickle_load(p):
    """Loads an object from a pickle file.

    Args:
        p (str or Path): The file path from where the object will be loaded.

    Returns:
        any: The loaded object, or None if the file does not exist or an error occurred.
    """
    p = Path(p)
    p = p.with_suffix(".pickle")
    if not p.is_file():
        return None
    try:
        with open(p, "rb") as f:
            d = pickle.load(f)
        return d
    except Exception as exc:
        print(str(exc))
        return 1


def np_save(p, d, allow_pickle=True):
    """Saves a NumPy array to a .npy file.

    Args:
        p (str or Path): The file path where the array will be saved.
        d (np.ndarray): The NumPy array to save.
        allow_pickle (bool): If True, allows saving of object arrays.

    Returns:
        int: 0 if successful, 1 if an error occurred.
    """
    p = Path(p)
    p = p.with_suffix(".npy")
    try:
        np.save(p, d, allow_pickle=allow_pickle)
        return 0
    except Exception as exc:
        print(str(exc))
        return 1


def np_load(p, allow_pickle=True):
    p = Path(p)
    p = p.with_suffix(".npy")
    if not p.is_file():
        return None
    try:
        if allow_pickle:
            d = np.load(p, allow_pickle=allow_pickle)
            if d is None:
                return None
            return d.item()
        else:
            return np.load(p)
    except Exception as exc:
        print(str(exc))
        return 1


async def np_load_async(p, allow_pickle=False):
    """Asynchronously loads a NumPy array from a .npy file.

    Args:
        p (str or Path): The file path from where the array will be loaded.
        allow_pickle (bool): If True, allows loading of object arrays.

    Returns:
        np.ndarray: The loaded array, or None if an error occurred.
    """
    p = Path(p)
    p = p.with_suffix(".npy")
    if not p.is_file():
        return None
    try:
        async with aiofiles.open(p, mode="rb") as f:
            buffer = await f.read()
        f = io.BytesIO(buffer)
        if allow_pickle:
            d = np.load(f, allow_pickle=allow_pickle)
            if d is None:
                return None
            return d.item()
        else:
            return np.load(f, allow_pickle=allow_pickle)
    except Exception:
        return None


def df_save(p, df):
    """Saves a DataFrame to a feather file.

    Args:
        p (str or Path): The file path where the DataFrame will be saved.
        df (pd.DataFrame): The DataFrame to save.

    Returns:
        int: 0 if successful, 1 if an error occurred.
    """
    p = Path(p)
    p = p.with_suffix(".feather")
    try:
        df.to_feather(p)
        return 0
    except Exception:
        return 1


def df_load(p):
    """Loads a DataFrame from a feather file.

    Args:
        p (str or Path): The file path from where the DataFrame will be loaded.

    Returns:
        pd.DataFrame: The loaded DataFrame, or None if an error occurred.
    """
    p = Path(p)
    p = p.with_suffix(".feather")
    if not p.is_file():
        return None
    try:
        df = pd.read_feather(p)
        return df
    except Exception:
        return None


def json_save(p, d):
    """Saves an object to a JSON file.

    Args:
        p (str or Path): The file path where the JSON will be saved.
        d (dict): The dictionary to save.

    Returns:
        int: 0 if successful, 1 if an error occurred.
    """
    p = Path(p).with_suffix(".json")
    try:
        with open(p, "wb") as f:
            f.write(orjson.dumps(d, option=orjson.OPT_NON_STR_KEYS))
        return 0
    except Exception as exc:
        print(f"json_save: file={p}, Exception: {exc}")
        return 1


def json_load(p):
    """Loads a JSON file into a dictionary.

    Args:
        p (str or Path): The file path from where the JSON will be loaded.

    Returns:
        dict: The loaded JSON as a dictionary, or None if an error occurred.
    """
    p = Path(p).with_suffix(".json")
    if not p.is_file():
        print(f"json_load: file {p} not found")
        return None
    try:
        with open(p, "rb") as f:
            return orjson.loads(f.read())
    except Exception as exc:
        print(f"json_load: file={p}, Exception: {exc}")
        return None

def json_dumps(o):
    if o is None:
        return None
    try:
        return orjson.dumps(o, option=orjson.OPT_SERIALIZE_NUMPY).decode()
    except Exception as exc:
        print(f"json_dumps Exception: {exc}")
        return 1

def json_loads(s):
    if s is None or s == "":
        return None
    try:
        o = orjson.loads(s)
        if o == {}:
            return None
        return o
    except:
        return None

# To be reviwed (used in PW) and renamed to json_load
def read_json(p):
    return json_load(p)

# ################################################################################
# Shell commands: cp, rm, sh_run, run_asyc
# ################################################################################


def cp(p1, p2):
    """Copies a file or directory to a new location.

    Args:
        p1 (str or Path): The source file or directory.
        p2 (str or Path): The destination path.

    Returns:
        int: 0 if successful, 1 if an error occurred.
    """
    p1, p2 = Path(p1), Path(p2)
    if not p1.exists():
        print(f"File {p1} not found")
        return 1
    try:
        if p1.is_file():
            shutil.copyfile(p1, p2)
            return 0
        if p2.is_dir():
            shutil.rmtree(p2)
        shutil.copytree(p1, p2)
    except Exception:
        return 1


def rm(p):
    """Removes a file or directory.

    Args:
        p (str or Path): The file or directory to remove.

    Returns:
        int: 0 if successful, 1 if an error occurred.
    """
    try:
        p = Path(p)
        if p.is_file():
            p.unlink()
            return 0
        if p.is_dir():
            return shutil.rmtree(p)
    except Exception:
        return 1


def sh_run(cmd):
    """Executes a shell command and returns the output.

    Args:
        cmd (str): The command to execute.

    Returns:
        tuple: A tuple containing the exit code and the command's stdout.
    """
    proc = subprocess.Popen(
        cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True
    )
    (stdout, stderr) = proc.communicate()
    stdout, stderr = stdout.decode("utf-8"), stderr.decode("utf-8")
    if proc.returncode != 0:
        raise RuntimeError(
            f"sh_run command \n{cmd}\n exit with error code {proc.returncode}:\n{stderr}"
        )
    return (proc.returncode, stdout)


async def run_asyc(*args):
    """Asynchronously executes a shell command and returns the output.

    Args:
        *args: The command and arguments to execute.

    Returns:
        tuple: A tuple with the exit code, stdout, and stderr.
    """
    try:
        process = await asyncio.create_subprocess_exec(
            *args, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
        )
        stdout, stderr = await process.communicate()
        return process.returncode, stdout.decode(), stderr.decode()
    except:
        return -1, '', 'Command not found'

# ##############################################################################################################
# upload_file, download_file
# ##############################################################################################################

def upload_file(url, p):
    """
    Uploads a file to the given pre-signed URL.
    Args:
        url (str): The pre-signed URL for the upload.
        p (str): The local path to the file to be uploaded.
    Returns 0 if upload is successful, 1 otherwise.
    """
    headers = {'Content-Type': 'application/octet-stream'}
    try:
        with open(p, 'rb') as f:
            res = requests.put(url, data=f, headers=headers)
            if res.status_code == 200: return 0
            print(f'upload_file ERROR: {res.status_code}:{res.text}')
            return 1
    except Exception as exc:
        print(f'upload_file EXCEPTION: {exc}')
        return 1
    
def download_file(url, p):
    """
    Downloads a file to the given pre-signed URL.
    Args:
        url (str): The pre-signed URL for the upload.
        p (str): The local path to the file to be uploaded.
    Returns 0 if upload is successful, 1 otherwise.
    """
    headers = {'Content-Type': 'application/octet-stream'}
    try:
        res = requests.get(url, headers=headers)
        if res.status_code==200:
            with open(p, 'wb') as f:
                f.write(res.content)
            return 0
        print(f'download_file ERROR: {res.status_code}:{res.text}')
        return 1
    except Exception as exc:
        print(f'download_file EXCEPTION: {exc}')
        return 1
    
# ##########################################################################################
# download_file, download_files
# ##########################################################################################
    
def download_file_by_subprocess(url, download_dir):
    """Downloads a single file using wget subprocess to specified directory. Returns 0 on success, 1 on failure."""
    cmd = shlex.split(f'wget -q -P {download_dir} {url}')
    try:
        with open('/tmp/download_file_by_subprocess.output', 'w') as f:
            completed_process = subprocess.run(cmd, stdout=f, check=True)
        # completed_process = subprocess.run(cmd, check=True)
        return 0
    except subprocess.CalledProcessError as e:
        return 1
    
# def download_files(urls, download_dir):
#     """Downloads multiple files in parallel using ProcessPoolExecutor. Returns 0 on success, 1 on failure."""
#     with concurrent.futures.ProcessPoolExecutor() as executor:
#         futures = {executor.submit(download_file_by_subprocess, url, download_dir): url for url in urls}
#         for future in concurrent.futures.as_completed(futures):
#             url = futures[future]
#             try:
#                 result = future.result()
#             except Exception as e:
#                 print(e)
#                 return 1
#             else:
#                 print(f'Download completed: {url}')
#     return 0

def download_files(urls, download_dir):
    """Downloads multiple files in parallel using ProcessPoolExecutor. Returns 0 on success, 1 on failure."""
    ctx = mp.get_context('spawn')
    with concurrent.futures.ProcessPoolExecutor(mp_context=ctx) as executor:
        futures = {executor.submit(download_file_by_subprocess, url, download_dir): url for url in urls}
        for future in concurrent.futures.as_completed(futures):
            url = futures[future]
            try:
                result = future.result()
            except Exception as e:
                print(e)
                return 1
            else:
                print(f'Download completed: {url}')
    return 0
    
# ################################################################################
# get_all_files, split_files_into_subfolders, get_dir_files, get_dir_dirs
# ################################################################################


def get_all_files(p, filetypes=None):
    """Returns all files of specified types in a directory and its subdirectories.

    Args:
        p (str or Path): The directory path to search.
        filetypes (list of str, optional): List of file extensions to filter by.

    Returns:
        list: A list of Path objects for each file found.
    """
    p = Path(p)
    filetypes = [e.lower() for e in listify(filetypes)]
    filetypes = [f".{e}" if e[0] != "." else e for e in filetypes]
    files = [p1 for p1 in p.rglob("*")]
    files = [
        e
        for e in files
        if xre("(DS_Store)", str(e)) is None and xre("(__MACOSX)", str(e)) is None
    ]
    if len(filetypes):
        files = [f for f in files if f.suffix.lower() in filetypes]
    return files


def split_files_into_subfolders(p, suffix=None, n=1000):
    """Splits files in a folder into subfolders with a maximum number of files each.

    Args:
        p (str or Path): The directory containing files to split.
        suffix (str, optional): File extension to filter by.
        n (int): Maximum number of files per subfolder.
    """
    p0 = Path(p)
    print('Collecting all files')
    ps = get_all_files(p0, suffix)
    print(f'Found {len(ps)} files')
    n = math.ceil(len(ps)/1000)
    for k in tqdm(range(n)):
        (p0/f'{k}').mkdir(parents=True, exist_ok=True, mode=0o755)
        for i in range(k*1000,min((k+1)*1000,len(ps))):
            p1 = ps[i]
            p2 = p0 / f"{k}" / p1.name
            _ = p1.rename(p2)


# Draft
def get_dir_files(path, types=None):
    return [
        f
        for f in path.glob("**/*")
        if f.is_file() and (types == None or f.suffix[1:] in types)
    ]


# Draft
def get_dir_dirs(path):
    return [d for d in path.glob("**/*") if d.is_dir()]


# ###################################################################################
# zip_get_registered_formats, zip_register_formats, iszip, dozip, unzip, unzip_tree
# ###################################################################################


def zip_get_registered_formats():
    """Retrieves all registered formats for unpacking files.

    Returns:
        list: A list of supported unpacking file formats.
    """
    return sum([e[1] for e in shutil.get_unpack_formats()], [])


def zip_register_formats():
    zips = zip_get_registered_formats()
    if ".7z" not in zips and ".hsz" not in zips:
        print("Registering [.7z,.hsz]")
        shutil.register_unpack_format("7zip", [".7z", ".hsz"], unpack_7zarchive)


def dozip(p, remove=False):
    """Creates a ZIP archive from a file or directory.

    Args:
        p (str or Path): The file or directory to archive.
        remove (bool, optional): If True, removes the original file or directory.

    Returns:
        int: 0 if successful, 1 if an error occurred.
    """
    p = Path(p)
    if not p.exists():
        print(f"dozip: file not found: {p}")
        return 1
    try:
        shutil.make_archive(p, "zip", p)
        if remove:
            if rm(p):
                return 1
        return 0
    except Exception:
        return 1


def iszip(p):
    zipext = zip_get_registered_formats()
    archext = [".rar"]
    return (Path(p)).suffix in zipext + archext


def unzip(p, remove=False):
    """Unzips or untars a file.

    Args:
        p (str or Path): The archive file to unzip.
        remove (bool, optional): If True, removes the original archive file.

    Returns:
        int: 0 if successful, 1 if an error occurred.
    """
    p = Path(p)
    if not p.exists():
        print(f"unzip: file not found: {p}")
        return 1
    if p.is_dir():
        return 0
    if not iszip(p):
        return 0
    if p.suffix in [".rar"]:
        code, stdout = sh_run(f"unrar x {p} {p.parent}")
        if code != 0:
            print(f"unrar error: {stdout}")
            return 1
        if remove:
            if rm(p):
                return 1
        print(f"File {p} extracted from archive")
        return 0
    p_dir = p.with_suffix("")
    try:
        shutil.unpack_archive(p, p_dir)
        if remove:
            if rm(p):
                return 1
        return 0
    except Exception as exc:
        print(f"Exception: {exc}")
        return 1


def unzip_tree(root_path):
    """
    Recursively unzips all files in a given root path.
    If the root path is a zip file, it will unzip it.
    For directories, it will recursively process all files and subdirectories.
    """
    if root_path.is_file() and root_path.suffix == '.zip':
        extract_to = root_path.with_suffix('')
        unzip(root_path)
        root_path.unlink()
        unzip_tree(extract_to)
    elif root_path.is_dir():
        for item in root_path.iterdir():
            unzip_tree(item)

# ###################################################################################
# get_seq_path
# ###################################################################################

def get_seq_path(seq, path, pre='', suf='', levels=2):
    """
    Generates paths for files, converting a seq number into a path within a hierarchy of folders.
    Examples:
        1000: /000/000/000/000000000001
        1000: 000/000/001/000000001000
        999999999999: 999/999/999/999999999999
    Allways uses 3 hierarchy levels with 1000 elements each, converting seq up to 1 trillion
    (3 levels plus 1.000 files inside last level equals 1000**4 = 1 trilion).
    Includes base_path in the path as a prefix, and prefix pre and suffix suf in the file name.
    Automatically creates the subdirs if they dont exist.
    Returns the generated path.
    """
    assert seq < 1_000_000_000_000
    if levels==3:
        d = f'{seq:012}'
        p = Path(f'{str(path)}/{d[:3]}/{d[3:6]}/{d[6:9]}')
    else:
        d = f'{seq:09}'
        p = Path(f'{str(path)}/{d[:3]}/{d[3:6]}')
    if not p.is_dir(): p.mkdir(parents=True, exist_ok=True)
    return f'{p}/{pre}{d}{suf}'



"""
Google Storage (GS)
https://cloud.google.com/storage/docs/samples/storage-generate-signed-url-v4#storage_generate_signed_url_v4-python
"""

import os
import datetime
import re
import pickle
import mimetypes
import requests
from google.cloud import storage
from dl2050utils.core import oget
from dl2050utils.env import config_load
from dl2050utils.fs import json_save
import hashlib
import hmac
import urllib.parse
from pathlib import Path


class _URLSigner:
    """Internal HMAC-based URL signer for local backend."""

    def __init__(self, secret_key: str, base_url: str):
        self.secret = secret_key.encode("utf-8")
        self.base_url = base_url.rstrip("/")

    def _make_signature(self, method: str, bucket: str, blob: str, exp: int, max_size: int | None):
        payload = f"{method}\n{bucket}\n{blob}\n{exp}\n{max_size or ''}"
        return hmac.new(self.secret, payload.encode("utf-8"), hashlib.sha256).hexdigest()

    def generate_url(self, path: str, method: str, bucket: str, blob: str,
                     timeout: int, max_size: int | None = None) -> str:
        import time
        exp = int(time.time()) + timeout
        sig = self._make_signature(method, bucket, blob, exp, max_size)
        query = {
            "bucket": bucket,
            "blob": blob,
            "exp": exp,
            "method": method,
            "sig": sig,
        }
        if max_size is not None:
            query["max_size"] = max_size
        qs = urllib.parse.urlencode(query)
        return f"{self.base_url}{path}?{qs}"

class GS:
    """
    Google Cloud Storage helper class to manage buckets, files, and URLs.
    This class is a simplfied drive to interact with Google Storage, providing an abstraction
    to isolate the application level code from the underlying storage provider.
    If facilitates operations such as creating/deleting buckets, uploading/downloading files,
    generating signed URLs for blob uploads/downloads, and managing files in-memory or on disk.
    Attributes:
    default_location (str): Default location for creating GCS buckets.
    gc (storage.Client): Google Cloud storage client instance.
    """

    def __init__(self, service, default_location="europe-west1"):
        cfg = config_load(service)
        # Try Google Cloud first
        key_dict = oget(cfg, ["gcloud", "gs_key"])
        fs_cfg = oget(cfg, ["fs"]) or {}
        self.internal_token = fs_cfg.get("internal_token")
        if self.internal_token:
            os.environ["FS_INTERNAL_TOKEN"] = self.internal_token
        self.mode = None  # 'gcloud' or 'local'

        if key_dict is not None:
            # ---------- GCS MODE ----------
            assert key_dict["type"] == "service_account"
            credentials_p = "./gs-keyfile.json"
            json_save(credentials_p, key_dict)
            os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = credentials_p

            self.default_location = default_location
            self.gc = storage.Client()
            self.mode = "gcloud"
        elif fs_cfg.get("backend") == "local":
            # ---------- LOCAL MODE ----------
            self.mode = "local"
            self.default_location = "local"
            self.gc = None  # not used

            self.root_dir = Path(fs_cfg.get("root_dir", f"/data/{service}/fs"))
            self.root_dir.mkdir(parents=True, exist_ok=True)

            base_url = fs_cfg.get("url", "http://localhost:8001")
            secret = fs_cfg.get("secret")
            if not secret:
                raise RuntimeError("GS local backend enabled but fs.secret not configured")

            self._signer = _URLSigner(secret_key=secret, base_url=base_url)
        else:
            raise RuntimeError("GS: neither gcloud.gs_key nor fs.backend=local configured")

    # ####################################################################################################
    # Admin
    # ####################################################################################################

    def create_bucket(self, bucket_name):
        """
        Creates a new bucket in Google Cloud Storage.
        Args:
            bucket_name (str): The name of the bucket to create.
        Returns:
            int: 0 if the bucket was created successfully, 1 otherwise.
        """
        try:
            bucket = self.gc.bucket(bucket_name)
            bucket.storage_class = "STANDARD"
            new_bucket = self.gc.create_bucket(bucket, location=self.default_location)
            print(f"Bucket {new_bucket.name} created in {new_bucket.location}, storage class {new_bucket.storage_class}")
            return 0
        except Exception as exc:
            print(f"create_bucket EXCEPTION: {str(exc)}")
            return 1

    def remove_bucket(self, bucket_name, force_delete=False):
        """
        Deletes a bucket from Google Cloud Storage.
        Args:
            bucket_name (str): The name of the bucket to delete.
            force_delete (bool, optional): If True, deletes all objects in the bucket. Defaults to False.
        Returns:
            int: 0 if the bucket was deleted successfully, 1 otherwise.
        """
        try:
            bucket = self.gc.bucket(bucket_name)
            if force_delete:
                # Delete all objects in the bucket
                blobs = bucket.list_blobs()
                for blob in blobs:
                    blob.delete()
                print(f"All objects in bucket '{bucket_name}' have been deleted.")
            bucket.delete()
            print(f"Bucket '{bucket_name}' has been deleted.")
            return 0
        except Exception as exc:
            print(f"remove_bucket EXCEPTION: {str(exc)}")
            return 1

    def delete_object(self, bucket_name, blob_name):
        """
        Deletes a single object from the bucket, with exact match.
        Args:
            bucket_name (str): Name of the GCS bucket.
            blob_name (str): Name of the blob (object) to delete.
        Returns:
            int: False if the object was deleted successfully, True otherwise.
        """
        try:
            bucket = self.gc.bucket(bucket_name)
            blob = bucket.blob(blob_name)
            blob.delete()
            print(f"Deleted object '{blob_name}' from bucket '{bucket_name}'.")
            return 0
        except Exception as exc:
            print(f"delete_object EXCEPTION: {str(exc)}")
            return 1

    def delete_objects(self, bucket_name, pattern=None, prefix=None, suffix=None):
        """
        Deletes multiple objects from a bucket matching a pattern.
        Args:
            bucket_name (str): Name of the GCS bucket.
            pattern (str, optional): Regular expression pattern to match blob names.
            prefix (str, optional): Prefix string that blob names should start with.
            suffix (str, optional): Suffix string that blob names should end with.
        Returns:
            int: The number of objects deleted.
        """
        try:
            bucket = self.gc.bucket(bucket_name)
            # Decide which blobs to list based on prefix
            blobs = bucket.list_blobs(prefix=prefix) if prefix else bucket.list_blobs()
            deleted_count = 0
            pattern_compiled = re.compile(pattern) if pattern else None
            for blob in blobs:
                blob_name = blob.name
                match = True
                if pattern_compiled:
                    if not pattern_compiled.search(blob_name):
                        match = False
                if prefix:
                    if not blob_name.startswith(prefix):
                        match = False
                if suffix:
                    if not blob_name.endswith(suffix):
                        match = False
                if match:
                    blob.delete()
                    # print(f"Deleted object '{blob_name}' from bucket '{bucket_name}'.")
                    deleted_count += 1
            print(f"Deleted {deleted_count} objects from bucket '{bucket_name}'.")
            return deleted_count
        except Exception as exc:
            print(f"delete_objects EXCEPTION: {str(exc)}")
            return 0

    def list(self, bucket_name, subdir=""):
        """
        Lists all files in a specified subdirectory of a bucket.
        Args:
            bucket_name (str): Name of the GCS bucket.
            subdir (str, optional): Subdirectory path within the bucket. Defaults to '' (root).
        Returns:
            list: List of blob names in the specified subdirectory.
        """
        try:
            bucket = self.gc.bucket(bucket_name)
            blobs = bucket.list_blobs(prefix=subdir)
            file_list = [blob.name for blob in blobs if not blob.name.endswith("/")]
            print(f"Files in {bucket_name}/{subdir}: {file_list}")
            return file_list
        except Exception as exc:
            print(f"list_files EXCEPTION: {str(exc)}")
            return []

    # ###################################################################################################################
    # Memmory Download, Upload
    # ###################################################################################################################

    def upload_mem(self, bucket_name, blob_name, data,
                   content_type="application/octet-stream",
                   use_pickle=True):
        """
        Uploads data from memory to storage (GCS or local FS).
        """
        try:
            if use_pickle:
                data = pickle.dumps(data)
            elif isinstance(data, str):
                data = data.encode("utf-8")


            if self.mode == "gcloud":
                # --------- Google Cloud ---------
                bucket = self.gc.bucket(bucket_name)
                blob = bucket.blob(blob_name)
                blob.upload_from_string(data, content_type=content_type)
                return 0

            elif self.mode == "local":
                # --------- Local fs-server (fs.py) via HTTP ---------
                size = len(data)
                upload_url = self.upload_url(
                    bucket_name,
                    blob_name,
                    timeout=15 * 60,
                    size=size,
                )
                if not upload_url:
                    print("upload_mem (local) ERROR: could not use upload_url")
                    return 1

                resp = requests.put(
                    upload_url,
                    data=data,
                    headers={"Content-Type": content_type},
                    timeout=60,
                )

                if resp.status_code not in (200, 201):
                    print("upload_mem (local) ERROR:", resp.status_code, resp.text)
                    return 1

                return 0

            else:
                print("upload_mem ERROR: unknown mode", self.mode)
                return 1

        except Exception as exc:
            print(f"upload_mem EXCEPTION: {str(exc)}")
            return 1

    def download_mem(self, bucket_name, blob_name, as_string=False, encoding="utf-8", use_pickle=True):
        """
        Downloads a blob from the bucket into memory.
        Args:
            bucket_name (str): Name of the GCS bucket.
            blob_name (str): Name of the blob to download.
            as_string (bool, optional): If True, decodes the data using the specified encoding.
                                        Ignored if use_pickle is True. Defaults to False.
            encoding (str, optional): The encoding to use when decoding bytes to string. Defaults to 'utf-8'.
            use_pickle (bool, optional): If True, deserializes the data using pickle after downloading.
                                        Defaults to False.
        Returns:
            Any: The data from the blob, possibly decoded or deserialized.
        """
        try:
            if self.mode == "gcloud":
                # --------- Google Cloud ---------
                bucket = self.gc.bucket(bucket_name)
                blob = bucket.blob(blob_name)
                data = blob.download_as_bytes()

            elif self.mode == "local":
                # --------- Local fs-server (fs.py) via HTTP ---------
                download_url = self.download_url(
                    bucket_name,
                    blob_name,
                    timeout=24 * 3600,
                )
                if not download_url:
                    print("download_mem (local) ERROR: could not generate download_url")
                    return None
                internal_token = self.internal_token
                headers = {}
                if internal_token:
                    headers["X-Internal-Token"] = internal_token


                resp = requests.get(download_url, headers=headers, timeout=60)

                if resp.status_code != 200:
                    print("download_mem (local) ERROR:", resp.status_code, resp.text)
                    return None

                data = resp.content

            else:
                print("download_mem ERROR: unknown mode", self.mode)
                return None
                # Pós-processamento igual para os dois modos
            if use_pickle:
                data = pickle.loads(data)
            elif as_string:
                data = data.decode(encoding)
            return data
        except Exception as exc:
            print(f"download_mem EXCEPTION: {str(exc)}")
            return None
    # ###################################################################################################################
    # File Download, Upload
    # ###################################################################################################################

    def upload_file(
        self,
        bucket_name,
        blob_name,
        local_file_path,
        content_type="application/octet-stream",
    ):
        """
        Uploads a local file to a specified bucket and blob.
        - In GCS mode: uploads to Google Cloud.
        - In local mode: copies the file into root_dir / bucket / blob.
        """
        try:
            if self.mode == "gcloud":
                bucket = self.gc.bucket(bucket_name)
                blob = bucket.blob(blob_name)
                blob.upload_from_filename(local_file_path, content_type=content_type)
                return 0
            # LOCAL MODE
            dst = self.root_dir / bucket_name / blob_name
            dst.parent.mkdir(parents=True, exist_ok=True)
            import shutil
            shutil.copy2(local_file_path, dst)
            return 0
        except Exception as exc:
            print(f"upload_file EXCEPTION: {str(exc)}")
            return 1

    def download_file(self, bucket_name, blob_name, local_file_path):
        """
        Downloads a blob from the bucket to a local file.
        - In GCS mode: downloads from Google Cloud.
        - In local mode: copies from root_dir / bucket / blob.
        """
        try:
            if self.mode == "gcloud":
                bucket = self.gc.bucket(bucket_name)
                blob = bucket.blob(blob_name)
                blob.download_to_filename(local_file_path)
                return 0

            # LOCAL MODE
            elif self.mode == "local":
                # --------- Local fs-server (fs.py) via HTTP ---------
                download_url = self.download_url(
                    bucket_name,
                    blob_name,
                    timeout=24 * 3600,
                )
                if not download_url:
                    print("download_file (local) ERROR: could not generate download_url")
                    return 1

                internal_token = self.internal_token

                headers = {}
                if internal_token:
                    headers["X-Internal-Token"] = internal_token

                # stream to not load everyting in ram 
                with requests.get(download_url, headers=headers, stream=True, timeout=60) as r:
                    if r.status_code != 200:
                        print("download_file (local) ERROR:", r.status_code, r.text)
                        return 1

                    Path(local_file_path).parent.mkdir(parents=True, exist_ok=True)
                    with open(local_file_path, "wb") as f:
                        for chunk in r.iter_content(chunk_size=1024 * 1024):
                            if chunk:
                                f.write(chunk)

                return 0

            else:
                print("download_file ERROR: unknown mode", self.mode)
                return 1
        except Exception as exc:
            print(f"download_file EXCEPTION: {str(exc)}")
            return 1

    # ####################################################################################################
    # download_folder, upload_folder
    # ####################################################################################################

    def upload_folder(self, bucket_name, blob_name, local_folder_path):
        """
        Recursively uploads a local folder to the remote storage, preserving the directory structure.
        Args:
            bucket_name (str): Name of the remote bucket.
            blob_name (str, optional): Remote folder path within the bucket. Defaults to ''.
            local_folder_path (str): Path to the local folder to upload.
        """
        try:
            for root, dirs, files in os.walk(local_folder_path):
                for file in files:
                    # Full local file path
                    local_file_path = os.path.join(root, file)
                    # Compute relative path to maintain directory structure
                    relative_path = os.path.relpath(local_file_path, local_folder_path)
                    relative_path = relative_path.replace(
                        os.sep, "/"
                    )  # Ensure UNIX-style path separators
                    # Construct remote file path
                    remote_file_path = os.path.join(blob_name, relative_path).replace(
                        os.sep, "/"
                    )
                    # Upload the file
                    content_type, _ = mimetypes.guess_type(local_file_path)
                    if content_type is None:
                        content_type = "application/octet-stream"
                    not_success = self.upload_file(
                        bucket_name,
                        remote_file_path,
                        local_file_path,
                        content_type=content_type,
                    )
                    if not_success:
                        print(f"Failed to upload {local_file_path}")
        except Exception as exc:
            print(f"upload_folder EXCEPTION: {str(exc)}")

    def download_folder(self, bucket_name, blob_name, local_folder_path):
        """
        Recursively downloads a folder from the remote storage to a local directory, preserving the directory structure.
        Args:
            bucket_name (str): Name of the remote bucket.
            blob_name (str): Remote folder path within the bucket.
            local_folder_path (str): Local path where the folder will be downloaded.
        """
        try:
            bucket = self.gc.bucket(bucket_name)
            blobs = bucket.list_blobs(prefix=blob_name)
            for blob in blobs:
                # Skip if the blob is a directory placeholder
                if blob.name.endswith("/"):
                    continue
                # Compute relative path to maintain directory structure
                relative_path = os.path.relpath(blob.name, blob_name)
                relative_path = relative_path.replace(os.sep, "/")
                # Construct local file path
                local_file_path = os.path.join(local_folder_path, relative_path)
                # Ensure local directories exist
                local_dir = os.path.dirname(local_file_path)
                if not os.path.exists(local_dir):
                    os.makedirs(local_dir)
                # Download the file
                result = self.download_file(bucket_name, blob.name, local_file_path)
                if result != 0: print(f"Failed to download {blob.name}")
        except Exception as exc:
            print(f"download_folder EXCEPTION: {str(exc)}")

    # ####################################################################################################
    # Signed urls
    # ####################################################################################################

    def upload_url(self, bucket_name, blob_name, timeout=15 * 60, size=None):
        """
        Generates a signed URL for uploading a blob.
        - Local mode: signed URL for local fileserver (/upload).
        """
        if self.mode == "gcloud":
            try:
                bucket = self.gc.bucket(bucket_name)
                blob = bucket.blob(blob_name)
                query_parameters = (
                    None if size is None else {"x-goog-content-length-range": f"0,{size}"}
                )
                url = blob.generate_signed_url(
                    version="v4",
                    expiration=datetime.timedelta(seconds=timeout),
                    method="PUT",
                    content_type="application/octet-stream",
                    query_parameters=query_parameters,
                )
                return url
            except Exception as exc:
                print(f"upload_url EXCEPTION: {str(exc)}")
                return None

        # LOCAL MODE
        try:
            return self._signer.generate_url(
                path="/upload",
                method="PUT",
                bucket=bucket_name,
                blob=blob_name,
                timeout=timeout,
                max_size=size,
            )
        except Exception as exc:
            print(f"upload_url (local) EXCEPTION: {str(exc)}")
            return None

    def download_url(self, bucket_name, blob_name, timeout=24 * 3600):
        """
        Generates a signed URL for downloading a blob.
        - Local mode: signed URL for local fileserver (/download).
        """
        if self.mode == "gcloud":
            try:
                bucket = self.gc.bucket(bucket_name)
                blob = bucket.blob(blob_name)
                url = blob.generate_signed_url(
                    version="v4",
                    expiration=datetime.timedelta(seconds=timeout),
                    method="GET",
                )
                return url
            except Exception as exc:
                print(f"download_url EXCEPTION: {str(exc)}")
                return None

        # LOCAL MODE
        try:
            return self._signer.generate_url(
                path="/download",
                method="GET",
                bucket=bucket_name,
                blob=blob_name,
                timeout=timeout,
                max_size=None,
            ) 
        except Exception as exc:
            print(f"download_url (local) EXCEPTION: {str(exc)}")
            return None

    def urls(self, bucket_name, blob_name, timeout=24 * 3600, size=None):
        """
        Generates both upload and download signed URLs for a blob.
        """
        return (
            self.upload_url(bucket_name, blob_name, timeout=timeout, size=size),
            self.download_url(bucket_name, blob_name, timeout=timeout),
        )



import pandas as pd
import os

def list_prefixes(df: pd.DataFrame) -> list:
    """Return all distinct prefixes in the dataframe."""
    return df["prefix"].dropna().unique().tolist()


def filter_by_prefix(df: pd.DataFrame, prefix: str) -> pd.DataFrame:
    """Return all rows that match a given prefix exactly."""
    return df[df["prefix"] == prefix]


def filter_prefix_contains(df: pd.DataFrame, text: str) -> pd.DataFrame:
    """Return all rows where prefix contains the given text."""
    return df[df["prefix"].str.contains(text, na=False)]


def find_by_uid_suffix(df: pd.DataFrame, uid_suffix: str) -> pd.DataFrame:
    """Return all rows that match a given uid_suffix."""
    return df[df["uid_suffix"] == uid_suffix]


def find_by_uid_full(df: pd.DataFrame, uid_full: str) -> pd.DataFrame:
    """Return all rows that match a given uid_full."""
    return df[df["uid_full"] == uid_full]


def holter_only(df: pd.DataFrame) -> pd.DataFrame:
    """Return only rows where holter == True."""
    return df[df["holter"] == True]


def non_holter_only(df: pd.DataFrame) -> pd.DataFrame:
    """Return only rows where holter == False."""
    return df[df["holter"] == False]


def get_path_by_uid_suffix(df: pd.DataFrame, uid_suffix: str) -> str | None:
    """
    Return the path for a given uid_suffix.
    If there are multiple rows, returns the first one.
    If nothing is found, returns None.
    """
    rows = df[df["uid_suffix"] == uid_suffix]
    if rows.empty:
        return None
    return rows.iloc[0]["path"]


def get_paths_by_prefix(df: pd.DataFrame, prefix: str, holter_only_flag: bool | None = None) -> list:
    """
    Return a list of paths filtered by prefix and optionally holter flag.
    - holter_only_flag = True  → only holter rows
    - holter_only_flag = False → only non-holter rows
    - holter_only_flag = None  → ignore holter column
    """
    subset = df[df["prefix"] == prefix]
    if holter_only_flag is not None:
        subset = subset[subset["holter"] == holter_only_flag]
    return subset["path"].dropna().tolist()


def check_missing_files(df):
    """
    Return subset of rows whose 'path' does not point to an existing file.
    """
    mask = ~df["path"].astype(str).apply(os.path.exists)
    return df[mask]


def check_existing_files(df):
    """
    Return subset of rows whose 'path' exists.
    """
    mask = df["path"].astype(str).apply(os.path.exists)
    return df[mask]



# setup.py
from setuptools import setup
from dl2050utils.__config__ import (
    name,
    package,
    description,
    author,
    author_email,
    keywords,
    get_camel,
)
from dl2050utils.__version__ import version

version_camel = get_camel(version)

classifiers = [
    "Development Status :: 4 - Beta",
    "Intended Audience :: Developers",
    "Topic :: Software Development :: Build Tools",
    "License :: OSI Approved :: MIT License",
    "Programming Language :: Python :: 3",
]

setup(
    name=name,
    version=version,
    packages=[package],
    license="MIT",
    description=description,
    author=author,
    author_email=author_email,
    keywords=keywords,
    classifiers=classifiers,
)
